/*************************************************************************/
/*									 */
/* Anbindung von MySQL an Eclipse					 */
/* Dateiname: 	mysql.c		        Version 	 0.4		 */
/* Autor: 	Thomas H. Kolbe		Letzte nderung: 22.01.2003 */
/*									 */
/*************************************************************************/

#include <stdio.h>

/* MySql Includes */
#include <mysql/mysql.h>

/* Eclipse Includes */
#include <external.h>

/* ------------------------------- Definitionen ---------------------------- */

#define MAX_SELECT_COLS	30	/* max. Anzahl von Spalten bei select()	 */
#define MAX_CURSORS	30	/* max. Anzahl offener Cursor		 */
#define MAX_LEN_SQLCMD  1000    /* max. Lnge von SQL-Kommandos          */

/* ---------------------------- globale Variablen -------------------------- */

MYSQL           mysql;
int             Logged_in=0;
static short	Cursor_open_flag[MAX_CURSORS];
MYSQL_RES       *Current_results[MAX_CURSORS];
short	        Var_Anz[MAX_CURSORS];
short	        Var_Num[MAX_CURSORS][MAX_SELECT_COLS];

/* -------------------------------- Prozeduren ----------------------------- */
 
 int p_logout()         /* Log out from the database and disconnect */
{
    if (Logged_in) 
    {
	mysql_close(&mysql);
	Logged_in=0;
    }
    Succeed;	
}

void print_mysql_error()    /* MySQL error message */
{
    Fprintf(Current_Output,"\nMySQL error: %s\n",mysql_error(&mysql));
}

/* Logon to the server and connect to the database */

int p_login(value val1, type tag1, value val2, type tag2, 
            value val3, type tag3, value val4, type tag4)
{
    char *database,*user,*passwd,*host;
    int  i;

    switch(TagType(tag1))    /* database */
        {
	    case TDICT:
	        {
		    database=DidName(val1.did);
		    break;
		}
	    case TSTRG:
	    {
		database=StringStart(val1);
		break;
	    }
	    default:
	    {
		Error(TYPE_ERROR);
	    }
	}
    switch(TagType(tag2))    /* username */
        {
	    case TDICT:
	    {
		user=DidName(val2.did);
		break;
	    }
	    case TSTRG:
	    {
		user=StringStart(val2);
		break;
	    }
	    default:
	    {
		Error(TYPE_ERROR);
	    }
	}
    switch(TagType(tag3))    /* password */
        {
	    case TDICT:
	    {
		passwd=DidName(val3.did);
		break;
	    }
	    case TSTRG:
	    {
		passwd=StringStart(val3);
		break;
	    }
	    default:
	    {
		Error(TYPE_ERROR);
	    }
	}
    switch(TagType(tag4))    /* server address */
        {
	    case TDICT:
	    {
		host=DidName(val4.did);
		break;
	    }
	    case TSTRG:
	    {
		host=StringStart(val4);
		break;
	    }
	    default:
	    {
		Error(TYPE_ERROR);
	    }
	}

    /* Initialise MySQL and connect to server */

    if (!(Logged_in))
    {
	mysql_init(&mysql);
	if (mysql_real_connect(&mysql,host,user,passwd,database,0,NULL,0))
	{
	    Logged_in=1;
	    for (i=0; i<MAX_CURSORS; i++)
	        {
	        Cursor_open_flag[i]=0;
		Current_results[i]=NULL;
		}
	    Succeed;
	} else {
	    Fprintf(Current_Output,
		    "Failed to connect to database: Error: %s\n",
		    mysql_error(&mysql));
	    Fail;
	}
    } else {
	Fprintf(Current_Output,"Error in sql_login: already logged in!\n");
	Fail;
    }
}

int p_lastmsg(value val1, type tag1, value val2, type tag2, 
	      value val3, type tag3)
{
	value	     str;
	unsigned int returncode;
	my_ulonglong rows_processed;
	char         *msg;

	Prepare_Requests;
	if (!(IsRef(tag1) || IsInteger(tag1)))	/* MySQL Return-Code */
		Error(TYPE_ERROR);
	if (!(IsRef(tag2) || IsString(tag2)))	/* MySQL Return-Message */
		Error(TYPE_ERROR);
	if (!(IsRef(tag3) || IsInteger(tag3)))	/* Rows processed */
		Error(TYPE_ERROR);
	returncode=mysql_errno(&mysql);
	msg=mysql_error(&mysql);
	Cstring_To_Prolog(msg,str);
	rows_processed=mysql_affected_rows(&mysql);
	Request_Unify_Integer(val1,tag1,returncode);
	Request_Unify_String(val2,tag2,str.ptr);
	Request_Unify_Integer(val3,tag3,(int)rows_processed);
	Return_Unify;
}

int p_execsql(value val1, type tag1)
{
    char *str;
    
    switch(TagType(tag1))
    {
	case TDICT:
	{
	    str=DidName(val1.did);
	    break;
	}
	case TSTRG:
	{
	    str=StringStart(val1);
	    break;
	}
	default:
	{
	    Error(TYPE_ERROR);
	}
    }
#ifdef DEBUG
    Fprintf(Current_Output,"SQL: %s\n",str);
#endif
    mysql_query(&mysql,str);
    Succeed;	
}

int p_showtables()
{
    char      *str;
    MYSQL_RES *result;
    MYSQL_ROW row;
    unsigned int num_fields;
    unsigned int i;
    unsigned long *lengths;
    
    if (result=mysql_list_tables(&mysql,NULL))
    {
	num_fields = mysql_num_fields(result);
	while ((row = mysql_fetch_row(result)))
	{
	    lengths = mysql_fetch_lengths(result);
	    for(i = 0; i < num_fields; i++)
	    {
		Fprintf(Current_Output,"  %.*s", 
			(int) lengths[i], row[i] ? row[i] : "NULL");
	    }
	    Fprintf(Current_Output,"\n");
	}
	mysql_free_result(result);
	Succeed;	
    } 
    else
    {
	print_mysql_error();
	Cut_External;
	Fail;
    }
}

void close_cursor(int cursor_nr)    /* Schliesst einen offenen Select-*/
{                                   /* Cursor.			      */
    if (Current_results[cursor_nr] != NULL)
        mysql_free_result(Current_results[cursor_nr]);
    Current_results[cursor_nr]=NULL;
}

void free_cursor(int cursor_nr)     /* Schliesst einen offenen Select- */
{				    /* Cursor und gibt die dazu        */
    close_cursor(cursor_nr);        /* reservierten Speicherbloecke    */
    Cursor_open_flag[cursor_nr]=0;  /* wieder frei. 		       */
}

void p_selectclosecursor(value val, type tag)  /* Diese Prozedur wird von   */
{                                   /* Eclipse automatisch aufgerufen, wenn */
#ifdef DEBUG
    Fprintf(Current_Output,"p_selectclosecursor(%d) called\n",val.nint);
#endif
    free_cursor(val.nint);          /* ein cut oder ein fail auftritt.	    */
}

int p_queryopencursor(value val1, type tag1, value val7, type tag7, 
		      value val2, type tag2, value val3, type tag3, 
		      value val4, type tag4, value val6, type tag6, 
		      value val5, type tag5)
{
    pword cursor2;
    pword *pw1,*pw2, *pw3, *pw4,*pw5, *pw6;
    char  where_clause[400], attributelist[400], hilfstring[80], distinct[10];
    char  sqlcmd[MAX_LEN_SQLCMD];
    char  *str1,*str3,*str4,*str6;
    short attrlist_flag=0, where_clause_flag=0, anz_variablen=0;
    short attrcount, varcount;
    int	  i, j, cursor_nr;
    MYSQL_RES *myresult;

    /* Ueberpruefung der Parameter */
    /* tag1 muesste noch auf uninstanziierte Variable getestet werden */
    Check_List(tag2);
    Check_List(tag5);
    Check_Integer(tag7);	/* DISTINCT-Flag darf nur 0 oder 1 sein */
    switch(TagType(tag3))
    {
	case TDICT:
	{
	    str3=DidName(val3.did);
	    break;
	}
	case TSTRG:
	{
	    str3=StringStart(val3);
	    break;
	}
	default:
	{
	    Error(TYPE_ERROR);
	}
    }
    switch(TagType(tag4))
    {
	case TDICT:
	{
	    str4=DidName(val4.did);
	    break;
	}
	case TSTRG:
	{
	    str4=StringStart(val4);
	    break;
	}
	default:
	{
	    Error(TYPE_ERROR);
	}
    }
    switch(TagType(tag6))
    {
	case TDICT:
	{
	    str6=DidName(val6.did);
	    break;
	}
	case TSTRG:
	{
	    str6=StringStart(val6);
	    break;
	}
	default:
	{
	    Error(TYPE_ERROR);
	}
    }

    cursor_nr=0;
    while ((cursor_nr<MAX_CURSORS) && (Cursor_open_flag[cursor_nr]))
	cursor_nr++;
    if (cursor_nr >= MAX_CURSORS)
    {
	Fprintf(Current_Output,
		"\nError: Cursor limit exhausted (max. no. is %d)!\n",
		MAX_CURSORS);
	Cut_External;
	Fail;
    }

    if (IsNil(tag2))
    {
	Fprintf(Current_Output,
		"\nError: attributelist must not be empty!\n");
	Cut_External;	/* nur Fail alleine funktioniert nicht */
	Fail;
    }
    if (IsNil(tag5))
    {
	Fprintf(Current_Output,
		"\nError: varlist must not be empty!\n");
	Cut_External;	/* nur Fail alleine funktioniert nicht */
	Fail;
    }

    /* Zunaechst die Variablenliste nach bereits instanziierten  */
    /* Variablen durchsehen und ueberpruefen, ob die */
    /* Attributliste (val2/tag2) nur aus bereits instanziierten */
    /* String-Werten besteht. */

    where_clause[0]='\0';
    attributelist[0]='\0';
    varcount=0;
    attrcount=0;

    pw1=val2.ptr;
    pw2=pw1+1;
    Dereference(pw1); 
    pw4=val5.ptr;
    pw5=pw4+1;
    Dereference(pw4);
    do
    {
	switch(TagType(pw1->tag))
	{
	    case TDICT:
	    {
		str1=DidName(pw1->val.did);
		break;
	    }
	    case TSTRG:
	    {
		str1=StringStart(pw1->val);
		break;
	    }
	    default:
	    {
		Error(TYPE_ERROR);
	    }
	}

	/* Ueberpruefen, ob an der aktuellen Position in der */
	/* Variablenliste eine freie Variable steht. Wenn ja, */
	/* dann wird das Attribut mit in die Liste der zu */
	/* projizierenden Attribute des SELECT-Befehls */
	/* aufgenommen. Ansonsten wird ein Gleichheitszeichen */
	/* zwischen dem Attribut und dem Wert an den WHERE-Teil */
	/* angehaengt. */
	if (attrlist_flag)
	    strcat(attributelist,",");
	strcat(attributelist,str1);
	attrlist_flag=1;
	if (IsRef(pw4->tag))
	{
	    Var_Num[cursor_nr][attrcount]=varcount;
	    varcount++;
	}
	else
	{
	    if (where_clause_flag)
		strcat(where_clause," and ");
	    strcat(where_clause,str1);
	    strcat(where_clause,"=");
	    switch(TagType(pw4->tag))
	    {
		case TDBL:
		/* case TFLOAT: */ /* only for ECLiPSe older than 5.5 */
		{
		    sprintf(hilfstring,"%f",
			    FloatVal(pw4->val,pw4->tag));
		    strcat(where_clause,hilfstring);
		    break;
		} 
		case TINT:
		{
		    sprintf(hilfstring,
			    "%d",pw4->val.nint);
		    strcat(where_clause,hilfstring);
		    break;
		}
		case TDICT:
		{
		    strcat(where_clause,"'");
		    strcat(where_clause,
			   DidName(pw4->val.did));
		    strcat(where_clause,"'");
		    break;
		}
		case TSTRG:
		{
		    strcat(where_clause,"'");
		    strcat(where_clause,
			   StringStart(pw4->val));
		    strcat(where_clause,"'");
		    break;
		}
		default:
		{
		    Error(TYPE_ERROR);
		}
	    }
	    where_clause_flag=1;
	    Var_Num[cursor_nr][attrcount]=-1;
	}
	attrcount++;

	pw3=pw2;
	pw6=pw5;
	if (!(IsNil(pw2->tag)))			/* naechstes Attribut */
	{
	    pw1=pw2->val.ptr;
	    pw2=pw1+1;
	    Dereference(pw1);
	    if (!(IsNil(pw5->tag)))		/* naechste Variable */
	    {
		pw4=pw5->val.ptr;
		pw5=pw4+1;
		Dereference(pw4);
	    }
	    else
	    {
		/* Die Variablenliste ist zu kurz! */
		Fprintf(Current_Output,
			"\nError: varlist too short!\n");
		Cut_External;
		Fail;
	    }
	}
    }
    while (!(IsNil(pw3->tag)));

    /* Das letzte Attribut wurde behandelt. Nun wird getestet, ob auch das */
    /* Ende der Variablenliste erreicht ist. Falls nein => Fehler! */

    if (!(IsNil(pw6->tag)))
    {
	/* Die Attributliste ist zu kurz! */
	Fprintf(Current_Output,
		"\nError: attributelist too short!\n");
	Cut_External;
	Fail;
    }

    Var_Anz[cursor_nr]=attrcount;
    if (val7.nint)
    {
	strcpy(distinct,"distinct ");
    }
    else
    {
	strcpy(distinct,"");
    }

    sprintf(sqlcmd,"select %s%s from %s",distinct,attributelist,str3);

    if (where_clause_flag)
    {
	strcat(sqlcmd," where ");
	strcat(sqlcmd,where_clause);
	if (strlen(str4))
	{
	    strcat(sqlcmd," and (");
	    strcat(sqlcmd,str4);
	    strcat(sqlcmd,")");
	}
    }
    else
	if (strlen(str4))
	{
	    strcat(sqlcmd," where ");
	    strcat(sqlcmd,str4);
	}
    if (strlen(str6))
    {
	strcat(sqlcmd," ");
	strcat(sqlcmd,str6);
    }
#ifdef DEBUG
    Fprintf(Current_Output,"SQL: %s\n",sqlcmd);
#endif
    /* Zunaechst testen, ob der entsprechende Cursor nicht noch */
    /* geoeffnet ist. Wenn ja, dann diesen zunaechst schliessen */
    if (Cursor_open_flag[cursor_nr])
    {
        /****** KANN DAS HIER BERHAUPT PASSIEREN? ******/
        /* cursor_nr zeigt doch normalerweise auf einen */
        /* unbelegten Cursor?!?                         */
        /************************************************/
	free_cursor(cursor_nr);
    }
    if (mysql_query(&mysql,sqlcmd))
        goto sorry_no_cursor;
    if (mysql_errno(&mysql) > 0)
        goto sorry_no_cursor;
    myresult=mysql_store_result(&mysql);
    if (myresult==NULL)
        goto sorry_no_cursor;

    Current_results[cursor_nr]=myresult;
/*    numfields=mysql_num_fields(myresult);
    if (numfields != attrcount)
    {
        Fprintf(Current_Output,"\nError: attributelist and variablelist ");
	Fprintf(Current_Output,"must have the same length!\n");
	close_cursor(cursor_nr);
	Cut_External;
	Fail;
    }
*/
    Cursor_open_flag[cursor_nr]=1;
    Make_Integer(&cursor2,cursor_nr);
    schedule_cut_fail_action(p_selectclosecursor,cursor2.val,cursor2.tag);
    Return_Unify_Integer(val1,tag1,cursor_nr);

sorry_no_cursor:
    print_mysql_error();
    Cut_External;	/* nur Fail alleine funktioniert nicht */
    Fail;
}

int p_selectopencursor(value val1, type tag1, value val2, type tag2, 
                       value val3, type tag3)
{
    pword	 cursor2;
    pword	 *pw1,*pw2,*pw3;
    char	 hilfstring[80];
    char         sqlcmd[MAX_LEN_SQLCMD];
    char	 *str1;
    short	 attrcount;
    int	         cursor_nr,numfields;
    MYSQL_RES    *myresult;

    /* Parameter checking */
    /* tag1 muesste noch auf uninstanziierte Variable getestet werden */
    Check_List(tag2);
    Check_List(tag3);

    cursor_nr=0;
    while ((cursor_nr<MAX_CURSORS) && (Cursor_open_flag[cursor_nr]))
        cursor_nr++;
    if (cursor_nr >= MAX_CURSORS)
    {
	Fprintf(Current_Output,
		"\nError: Cursor limit exhausted (max. no. is %d)!\n",
		MAX_CURSORS);
	Cut_External;
	Fail;
    }
    
    if (IsNil(tag2))
    {
        Fprintf(Current_Output,
		"\nError: select statement must not be empty!\n");
	Cut_External;	/* nur Fail alleine funktioniert nicht */
	Fail;
    }
    if (IsNil(tag3))
    {
      Fprintf(Current_Output,
	      "\nError: varlist must not be empty!\n");
      Cut_External;	/* nur Fail alleine funktioniert nicht */
      Fail;
    }

    /* Zunaechst die Laenge der Variablenliste bestimmen */

    attrcount=0;
    pw1=val3.ptr;
    pw2=pw1+1;
    Dereference(pw1); 
    do
    {
        Var_Num[cursor_nr][attrcount]=attrcount;
	attrcount++;

	pw3=pw2;
	if (!(IsNil(pw2->tag)))			/* naechstes Variable */
	{
	    pw1=pw2->val.ptr;
	    pw2=pw1+1;
	    Dereference(pw1);
	}
    }
    while (!(IsNil(pw3->tag)));
	Var_Anz[cursor_nr]=attrcount;

    /* nun wird der komplette SELECT-Befehl aufgebaut */

    strcpy(sqlcmd,"select ");
    pw1=val2.ptr;
    pw2=pw1+1;
    Dereference(pw1); 
    do
    {
        switch(TagType(pw1->tag))
	{
	    case TDICT:
	    {
	        str1=DidName(pw1->val.did);
		break;
	    }
	    case TSTRG:
	    {
	        str1=StringStart(pw1->val);
		break;
	    }
	    case TINT:
	    {
	        sprintf(hilfstring,"%d",pw1->val.nint);
		str1=hilfstring;
		break;
	    }
	    case TDBL:
	    /* case TFLOAT: */ /* only for ECLiPSe older than 5.5 */
	    {
	        sprintf(hilfstring,
			"%lf",FloatVal(pw1->val,pw1->tag));
		str1=hilfstring;
		break;
	    }
	    default:
	    {
	        Error(TYPE_ERROR);
	    }
	}
	strcat(sqlcmd,str1);
	strcat(sqlcmd," ");

	pw3=pw2;
	if (!(IsNil(pw2->tag)))			/* naechstes Atom */
	{
	    pw1=pw2->val.ptr;
	    pw2=pw1+1;
	    Dereference(pw1);
	}
    }
    while (!(IsNil(pw3->tag)));
#ifdef DEBUG
    Fprintf(Current_Output,"SQL: %s\n",sqlcmd);
#endif
    /* Zunaechst testen, ob der entsprechende Cursor nicht noch */
    /* geoeffnet ist. Wenn ja, dann diesen zunaechst schliessen */
    if (Cursor_open_flag[cursor_nr])
    {
        /****** KANN DAS HIER BERHAUPT PASSIEREN? ******/
        /* cursor_nr zeigt doch normalerweise auf einen */
        /* unbelegten Cursor?!?                         */
        /************************************************/
        free_cursor(cursor_nr);
    }
    if (mysql_query(&mysql,sqlcmd))
        goto sorry_no_cursor;
    if (mysql_errno(&mysql) > 0)
        goto sorry_no_cursor;
    myresult=mysql_store_result(&mysql);
    if (myresult==NULL)
        goto sorry_no_cursor;

    Current_results[cursor_nr]=myresult;
    numfields=mysql_num_fields(myresult);
    if (numfields != attrcount)
    {
        Fprintf(Current_Output,"\nError: attributelist and variablelist ");
	Fprintf(Current_Output,"must have the same length!\n");
	close_cursor(cursor_nr);
	Cut_External;
	Fail;
    }
    Cursor_open_flag[cursor_nr]=1;
    Make_Integer(&cursor2,cursor_nr);
    schedule_cut_fail_action(p_selectclosecursor,cursor2.val,cursor2.tag);
    Return_Unify_Integer(val1,tag1,cursor_nr);

sorry_no_cursor:
    print_mysql_error();
    Cut_External;
    Fail;
}

int p_describeopencursor(value val1, type tag1, value val2, type tag2,
			 value val3, type tag3)
{
    pword	 cursor2;
    pword	 *pw1,*pw2,*pw3;
    char	 hilfstring[80];
    char         sqlcmd[MAX_LEN_SQLCMD];
    char	 *str;
    short	 attrcount;
    int	         cursor_nr,numfields;
    MYSQL_RES    *myresult;

    /* Parameter checking */
    /* tag1 muesste noch auf uninstanziierte Variable getestet werden */
    Check_List(tag3);
    switch(TagType(tag2))
    {
	case TDICT:
	{
	    str=DidName(val2.did);
	    break;
	}
	case TSTRG:
	{
	    str=StringStart(val2);
	    break;
	}
	default:
	{
	    Error(TYPE_ERROR);
	}
    }

    cursor_nr=0;
    while ((cursor_nr<MAX_CURSORS) && (Cursor_open_flag[cursor_nr]))
        cursor_nr++;
    if (cursor_nr >= MAX_CURSORS)
    {
	Fprintf(Current_Output,
		"\nError: Cursor limit exhausted (max. no. is %d)!\n",
		MAX_CURSORS);
	Cut_External;
	Fail;
    }
    
    if (IsNil(tag3))
    {
      Fprintf(Current_Output,
	      "\nError: varlist must not be empty!\n");
      Cut_External;	/* nur Fail alleine funktioniert nicht */
      Fail;
    }

    /* Zunaechst die Laenge der Variablenliste bestimmen */

    attrcount=0;
    pw1=val3.ptr;
    pw2=pw1+1;
    Dereference(pw1); 
    do
    {
        Var_Num[cursor_nr][attrcount]=attrcount;
	attrcount++;

	pw3=pw2;
	if (!(IsNil(pw2->tag)))			/* naechstes Variable */
	{
	    pw1=pw2->val.ptr;
	    pw2=pw1+1;
	    Dereference(pw1);
	}
    }
    while (!(IsNil(pw3->tag)));
	Var_Anz[cursor_nr]=attrcount;

    strcpy(sqlcmd,"describe ");
    strcat(sqlcmd,str);
#ifdef DEBUG
    Fprintf(Current_Output,"SQL: %s\n",sqlcmd);
#endif
    /* Zunaechst testen, ob der entsprechende Cursor nicht noch */
    /* geoeffnet ist. Wenn ja, dann diesen zunaechst schliessen */
    if (Cursor_open_flag[cursor_nr])
    {
        /****** KANN DAS HIER BERHAUPT PASSIEREN? ******/
        /* cursor_nr zeigt doch normalerweise auf einen */
        /* unbelegten Cursor?!?                         */
        /************************************************/
        free_cursor(cursor_nr);
    }
    if (mysql_query(&mysql,sqlcmd))
        goto sorry_no_cursor;
    if (mysql_errno(&mysql) > 0)
        goto sorry_no_cursor;
    myresult=mysql_store_result(&mysql);
    if (myresult==NULL)
        goto sorry_no_cursor;

    Current_results[cursor_nr]=myresult;
    numfields=mysql_num_fields(myresult);
    if (numfields != attrcount)
    {
        Fprintf(Current_Output,"\nError: attributelist and variablelist ");
	Fprintf(Current_Output,"must have the same length!\n");
	close_cursor(cursor_nr);
	Cut_External;
	Fail;
    }
    Cursor_open_flag[cursor_nr]=1;
    Make_Integer(&cursor2,cursor_nr);
    schedule_cut_fail_action(p_selectclosecursor,cursor2.val,cursor2.tag);
    Return_Unify_Integer(val1,tag1,cursor_nr);

sorry_no_cursor:
    print_mysql_error();
    Cut_External;
    Fail;
}

int p_selectfetch(value val1, type tag1, value val5, type tag5)
{
    pword         *pw4,*pw5;
    int	          i, cursor_nr;
    MYSQL_ROW     row;
    long          intnum;
    double        realnum;

    Prepare_Requests;

    /* Ueberpruefung der Parameter */
    Check_Integer(tag1);
    cursor_nr=val1.nint;
    if (cursor_nr >= MAX_CURSORS)
    {
        Fprintf(Current_Output,
		"\nError: cursor# too big (range is 0-%d)!\n",MAX_CURSORS-1);
	Cut_External;
	Fail;
    }

    /* Einen Wert auslesen, und, falls es keinen mehr gibt, den */
    /* Cursor wieder schliessen */

    if (row=mysql_fetch_row(Current_results[cursor_nr]))
    {
        /* Es wurde ein weiteres Tupel eingelesen */

        pw4=val5.ptr;
	pw5=pw4+1;
	Dereference(pw4);

	for (i=0; i<Var_Anz[cursor_nr]; i++)
	{
	    if (Var_Num[cursor_nr][i] != -1)
	    {
	        if (row[i]==NULL) /* Wert=NULL ? */
		{
		    Request_Unify_Atom(pw4->val,pw4->tag,Did("",0));
		}
		else
		{
#if MYSQL_VERSION_ID >= 32224
		    switch(mysql_fetch_field_direct(
			    Current_results[cursor_nr],i)->type)
#else
		    switch(mysql_fetch_field_direct(
			    Current_results[cursor_nr],i).type)
#endif
		    {
		    case FIELD_TYPE_VAR_STRING:  /* VARCHAR field */
		    case FIELD_TYPE_STRING:	 /* CHAR field */
		    case FIELD_TYPE_DATE:	 /* DATE field */
		    case FIELD_TYPE_DATETIME:    /* DATETIME field */
		      {
		        Request_Unify_Atom(pw4->val,pw4->tag,
					   Did((char *)row[i],0));
			break;
		      }
		    case FIELD_TYPE_TINY:	 /* TINY field */
		    case FIELD_TYPE_SHORT:	 /* SMALLINT field */
		    case FIELD_TYPE_LONG:	 /* INTEGER field */
		    case FIELD_TYPE_INT24:	 /* MEDIUMINT field */
		      {
			sscanf(row[i],"%ld",&intnum);
		        Request_Unify_Integer(pw4->val,
					      pw4->tag,
					      intnum);
			break;
		      }
		    case FIELD_TYPE_DECIMAL:     /* DECIMAL or NUMERIC field */
		    case FIELD_TYPE_FLOAT:	 /* FLOAT field */
		    case FIELD_TYPE_DOUBLE:	 /* DOUBLE or REAL field */
		      {
			sscanf(row[i],"%lf",&realnum);
		        Request_Unify_Float(pw4->val,
					    pw4->tag,
					    realnum);
			break;
		      }
		    default:                     /* Unknown datatypes are */
		      {                          /* handled like strings  */
		        Request_Unify_Atom(pw4->val,pw4->tag,
					   Did((char *)row[i],0));
		      }
		    }
		}
	    }
	    if (!(IsNil(pw5->tag)))
	    {
	        pw4=pw5->val.ptr;
		pw5=pw4+1;
		Dereference(pw4);
	    }			
	}
	Return_Unify; 
    } 
    else
    {
        /* Entweder ist ein Fehler aufgetreten oder es gibt kein    */
        /* Tupel mehr; das Praedikat muss als nicht mehr 	    */
        /* resatisfizierbar beendet werden. Durch das Fail wird die */
        /* Prozedur 'p_selectclosecursor' automatisch aufgerufen    */
        /* und der entsprechende Cursor geschlossen. 		    */

	Cut_External;
	Fail;
    }
}

